"""Make tables."""

from __future__ import annotations
from typing import Callable

import numpy as np
import pandas as pd

from risus.engine import target_number

def make_target_number_table(
        max_potency: int,
        n_faces: int = 6,
        **kwargs
) -> pd.DataFrame:

    """Make a dataframe comparing potencies to target numbers.

    This table includes all the target numbers up to the maximum possible for
    the highest potency (that is, 6*max_potency).

    # Arguments:
    * `max_potency`: The highest potency to consider.
    * `n_faces`: How many faces the dice should have.

    # Returns:
    A DataFrame whose columns are the potencies, whose rows are the target
    numbers, and whose contents are the probability (as a percentage) that a
    cliché of that potency would beat that target number.

    # Examples:
    >>> make_target_number_table(4)
            1      2      3      4
    1   100.0  100.0  100.0  100.0
    2    83.3  100.0  100.0  100.0
    3    66.7   97.2  100.0  100.0
    4    50.0   91.7   99.5  100.0
    5    33.3   83.3   98.1   99.9
    6    16.7   72.2   95.4   99.6
    7     0.0   58.3   90.7   98.8
    8     0.0   41.7   83.8   97.3
    9     0.0   27.8   74.1   94.6
    10    0.0   16.7   62.5   90.3
    11    0.0    8.3   50.0   84.1
    12    0.0    2.8   37.5   76.1
    13    0.0    0.0   25.9   66.4
    14    0.0    0.0   16.2   55.6
    15    0.0    0.0    9.3   44.4
    16    0.0    0.0    4.6   33.6
    17    0.0    0.0    1.9   23.9
    18    0.0    0.0    0.5   15.9
    19    0.0    0.0    0.0    9.7
    20    0.0    0.0    0.0    5.4
    21    0.0    0.0    0.0    2.7
    22    0.0    0.0    0.0    1.2
    23    0.0    0.0    0.0    0.4
    24    0.0    0.0    0.0    0.1

    >>> make_target_number_table(4, breakthrough=True)
            1      2      3      4
    1   100.0  100.0  100.0  100.0
    2    83.3  100.0  100.0  100.0
    3    66.7   97.2  100.0  100.0
    4    50.0   91.7   99.5  100.0
    5    33.3   83.3   98.1   99.9
    6    16.7   72.2   95.4   99.6
    7     0.0   58.3   90.7   98.8
    8     0.0   41.7   83.8   97.3
    9     0.0   27.8   74.1   94.6
    10    0.0   16.7   62.5   90.3
    11    0.0    8.3   50.0   84.1
    12    0.0    2.8   37.5   76.1
    13    0.0    0.0   25.9   66.4
    14    0.0    0.0   16.2   55.6
    15    0.0    0.0    9.3   44.4
    16    0.0    0.0    4.6   33.6
    17    0.0    0.0    1.9   23.9
    18    0.0    0.0    0.5   15.9
    19    0.0    0.0    0.0    9.7
    20    0.0    0.0    0.0    5.4
    21    0.0    0.0    0.0    2.7
    22    0.0    0.0    0.0    1.2
    23    0.0    0.0    0.0    0.4
    24    0.0    0.0    0.0    0.1

    """
    max_target_number = n_faces * max_potency

    tn_array = np.zeros((max_target_number, max_potency))
    for potency in range(1, max_potency+1):
        for tn in range(1, max_target_number+1):
            outcome = target_number(potency, tn, n_faces=n_faces, percent=True, **kwargs)
            tn_array[tn-1, potency-1] = round(outcome, 1)

    return pd.DataFrame(tn_array, range(1, max_target_number+1), range(1, max_potency+1))


def make_victory_table(
        max_potency: int,
        compare_func: Callable[..., float],
        **kwargs: bool
) -> pd.DataFrame:
    """Make a dataframe representing the chances that a given potency can beat another.

    # Arguments:
    *`max_potency`: The largest potency to go up against.
    *`compare_func`: The function to use to compare the two potencies, either
    `combat` or `single_action_conflict`.

    Also accepts keyword arguments as for `risus.engine`.

    Returns:
    --------
    A dataframe representing the chances that the row potency can beat the column potency in combat.

    Examples:
    ---------
    >>> from risus.engine import combat
    >>> make_victory_table(6, combat)
           1      2      3     4     5     6
    1   50.0    5.0    0.1   0.0   0.0   0.0
    2   95.0   50.0    8.2   0.3   0.0   0.0
    3   99.9   91.8   50.0  10.5   0.7   0.0
    4  100.0   99.7   89.5  50.0  12.2   1.1
    5  100.0  100.0   99.3  87.8  50.0  13.7
    6  100.0  100.0  100.0  98.9  86.3  50.0

    >>> from risus.engine import combat
    >>> make_victory_table(6, combat, inappropriate=True)
           1      2      3     4     5     6
    1   50.0   10.0    1.2   0.0   0.0   0.0
    2   95.0   52.5   16.4   3.5   0.3   0.0
    3   99.9   91.8   54.1  20.7   5.7   0.8
    4  100.0   99.7   89.7  55.2  23.9   7.7
    5  100.0  100.0   99.3  88.3  56.1  26.4
    6  100.0  100.0  100.0  98.9  87.1  56.8

    >>> from risus.engine import single_action_conflict
    >>> make_victory_table(8, single_action_conflict, n_faces=8)
           1      2     3     4     5     6     7     8
    1   50.0   11.6   1.7   0.2   0.0   0.0   0.0   0.0
    2   88.4   50.0  17.9   4.6   0.9   0.1   0.0   0.0
    3   98.3   82.1  50.0  22.0   7.4   2.0   0.5   0.1
    4   99.8   95.4  78.0  50.0  24.8   9.9   3.3   0.9
    5  100.0   99.1  92.6  75.2  50.0  27.0  12.0   4.6
    6  100.0   99.9  98.0  90.1  73.0  50.0  28.6  13.9
    7  100.0  100.0  99.5  96.7  88.0  71.4  50.0  30.0
    8  100.0  100.0  99.9  99.1  95.4  86.1  70.0  50.0

    """
    chance_of_victory = np.zeros((max_potency,max_potency))

    # This is a monstrous kludge and I hate the off-by-oneness of the whole situation.
    for potency_1, potency_2 in [(x,y) for x in range(1,max_potency+1) for y in range(1,max_potency+1)]:
        outcome = compare_func(potency_1, potency_2, percent=True, **kwargs)
        chance_of_victory[potency_1-1][potency_2-1] = round(outcome, 1)

    return pd.DataFrame(chance_of_victory, range(1,max_potency+1), range(1,max_potency+1))


def make_team_victory_table(
        enemy_potency: int,
        player_potency: int = 6,
        total_helpers: int = 4,
        **kwargs
) -> pd.DataFrame:
    """Make a victory table for a team fighting with damage_policy.

    This is a seperate function from make_victory_table because the team_combat
    function has a different arity than single_action_conflict and combat.

    This function runs in
    O(`enemy_potency*player_potency**2*total_helpers`), which is to
    say, really slowly. Caching doesn't really help, since all of the
    cells of the table need to be visited in any event.

    # Arguments:
    * `enemy_potency`: The largest potency to consider in the enemy axis.
    * `player_potency`: The strongest players to consider.
    * `total_helpers`: How many helpers to allow.

    Also accepts keyword arguments as for `risus.engine.team_combat`.

    # Returns:
    A DataFrame whose first axis represents the leader's potency and
    the second the helper's. The third axis is a Series representing
    the enemy's potency.

    # Examples:
    >>> from risus.damage_policy import damage_volunteer
    >>> from risus.engine import team_combat
    >>> team_combat(2, 4, 4, damage_volunteer)
    0.7055969748511421
    >>> make_team_victory_table(4, damage_policy=damage_volunteer)[2][4][4]
    0.7055969748511421

    """
    from risus.engine import team_combat

    # Whoosh...
    res = pd.DataFrame({leader_potency: {helper_potency:
                                              pd.Series({enemy_potency:
                                                         team_combat(leader_potency,
                                                                     helper_potency,
                                                                     enemy_potency,
                                                                     **kwargs)
                                                         for enemy_potency
                                                         in range(1, enemy_potency+1)})
                                              for helper_potency
                                              in range(1, player_potency*total_helpers+1)}
                             for leader_potency
                             in range(1, player_potency+1)})
    return res



"""The risus game engine.

This module implements the "unholy trinity" of the Risus game engine:
`combat`, `single_action_conflict`, and `target_number` checks. There
is also a seperate function for `team_combat`.

These functions all support breathrough rolls as described on page 54
of the Risus companion, as well as old-fashined exploding d6s.

"""
from __future__ import annotations
from functools import cache
from typing import Callable

from icepool import Die, d, reduce, Reroll


def combat(attack_potency: int,
           enemy_potency: int,
           inappropriate: bool = False,
           percent: bool = False,
           n_faces: int = 6,
           breakthrough: bool = False,
           explode: bool = False
) -> float:
    """Simulate basic Risus combat between two clichés of the given potencies.

    In Risus, two characters face off in combat by rolling against
    their chosen cliché: the lower roller loses one of their dice for
    the rest of the combat. In the case that the cliché that rolled
    higher is deemed "inappropriate" by the GM, then the low roller
    loses three die. See Risus page 2 for more detail.

    # Arguments:
    * `attack_potency`: The attacking cliché's potency; the only one
    that can be inappropriate in this function.
    * `enemy_potency`: The defending cliché's potency; this one can't be inappropriate.
    * `inappropriate`: Whether the first cliche is inappropriate.
    * `percent`: Whether to return the value as a percent.
    * `n_faces`: The number of faces on the dice.
    * `breakthrough`: Whether or not to follow breakthrough rules from the Risus companion.
    * `explode`: Whether or not to roll with exploding dice.

    # Returns:
    The chance (potentially as a percent) that the attacker beats the enemy.

    # Examples:
    >>> round(combat(4,3), 3)
    0.895

    """
    outcome = _combat(attack_potency, enemy_potency, explode=explode,
                      inappropriate=inappropriate, n_faces=n_faces, breakthrough=breakthrough)
    return outcome.probabilities(percent=percent)[True]


# This is fancy enough to deserve to be its own procedure, even though in the
# Risus rules it's considered a special case of ordinary combat.
def team_combat(
        leader_potency: int,
        helper_potency: int,
        enemy_potency: int,
        damage_policy: Callable[[int, int], tuple[int, int]],
        n_faces: int = 6,
        percent: bool = False,
        inappropriate: bool = False,
        breakthrough: bool = False,
        explode: bool = False
) -> float:
    """Simulate team combat.

    This procedure simply counts leader death as loss without attempting to
    reform the team and does not implement the double-damage self-sacrifice.

    # Arguments:
    * `leader_potency`: The potency of the team leader's cliché.
    * `helper_potency`: The total potency of the helpers' clichés.
    * `enemy_potency`: The potency of the enemy's cliché.
    * `damage_policy`: A function that takes the leader's and helpers' potencies
      and applies damage to them, returning a new pair of potencies.
    * `n_faces`: The number of faces on the dice.
    * `percent`: Whether or not to return the probability as a percent.
    * `inappropriate`: Whether or not all the team's clichés are inappropriate.
    * `breakthrough`: Whether or not to follow breakthrough rules from the Risus companion.
    * `explode`: Whether or not to roll with exploding dice.

    # Returns:
    The probability (potentially as a percentage) that the team is victorious.

    # Examples:
    >>> from risus.damage_policy import damage_team_mates_only
    >>> team_combat(4,0,4, damage_team_mates_only)
    0.5
    """
    outcome = _team_combat(leader_potency, helper_potency,
                           enemy_potency, damage_policy,
                           inappropriate=inappropriate, n_faces=n_faces,
                           breakthrough=breakthrough, explode=explode)

    return outcome.probabilities(percent=percent)[1]


def single_action_conflict(
        attack_potency: int,
        enemy_potency: int,
        n_faces: int = 6,
        percent: bool = False,
        breakthrough: bool = False,
        explode: bool = False,
) -> float:
    """Compute the chances of victory in a single-action conflict.

    The winner of a single action conflict is simply the higher roller. See
    Risus page 3.

    # Arguments:
    * `attack_potency`: The potency of the cliché whose chance of victory to compute.
    * `enemy_potency`: The potency of the cliché they're up against.
    * `percent`: Whether or not to return the value as a percent.
    * `n_faces`: The number of faces on the dice.
    * `breakthrough`: Whether to use breakthrough rules from the Risus Companion, page 54.
    * `explode`: Whether the attacker has exploding dice.

    # Returns:
    The probability (potentially as a percent) that the attacker is victorious.

    # Examples:
    >>> round(single_action_conflict(4, 3, percent=True), 1)
    79.5

    >>> single_action_conflict(1,6)
    0.0
    """
    attack_die = d(n_faces).explode() if explode else d(n_faces)
    attack_pool = (
        breakthrough_pool(attack_potency, attack_die) if breakthrough else attack_potency @ attack_die
    )

    enemy_die = d(n_faces)
    enemy_pool = enemy_potency @ enemy_die

    res_die = reduce(lambda a,b: Reroll if a == b else a > b, [attack_pool, enemy_pool]).simplify()
    # Catch a weird corner case where there's an automatic victory and so the
    # return is ill-formed:
    if attack_potency >= n_faces * enemy_potency:
        res_die = Die({True: 1, False: 0})

    if enemy_potency >= n_faces * attack_potency:
        res_die = Die({True: 0, False: 1})

    return res_die.probabilities(percent=percent)[1]


def target_number(
        potency: int,
        difficulty: int,
        n_faces: int = 6,
        percent: bool = False,
        explode: bool = False,
        breakthrough: bool = False
) -> float:
    """Compute the probability that a cliché with this potency will beat the target difficulty.

    To beat the target number the roll must be equal or greater than
    the difficulty: the rules are explained on Risus page 1.

    # Arguments:
    * `potency`: The potency of the cliché being rolled against.
    * `difficulty`: The target number to beat.
    * `percent`: Whether to return the value as a percentge rather
      than a number between 0 and 1.
    * `n_faces`: The number of faces for the dice.
    * `explode`: Whether to roll exploding dice.
    * `breakthrough`: Whether to use breakthroughs.

    # Returns:
    The probability that a cliché of this potency beats the target number.

    # Examples:
    >>> target_number(3, 10)
    0.625

    """
    die = d(n_faces).explode() if explode else d(n_faces)
    pool = breakthrough_pool(potency, die) if breakthrough else potency @ die

    res_die = pool >= difficulty
    # Catch a weird corner case where there's an automatic victory and so the
    # return is ill-formed:
    if potency >= difficulty:
        # Auto-success.
        res_die = Die({True: 1, False: 0})
    if difficulty > 6 * potency:
        # Auto-failure
        res_die = Die({True: 0, False: 1})

    return res_die.probabilities(percent=percent)[1]


@cache
def _combat(
        attack_potency: int,
        enemy_potency: int,
        n_faces: int = 6,
        inappropriate: bool = False,
        explode: bool = False,
        breakthrough: bool = False
) -> Die:
    """Internal combat calculator.

    # Returns:
    Die where True is the number of victorious cases and False the number of defeats.

    """
    attack_die = d(n_faces).explode() if explode else d(n_faces)
    enemy_die = d(n_faces)
    damage = 3 if inappropriate else 1

    # Base cases.
    if attack_potency <= 0 and enemy_potency > 0:
        # 2 wins.
        return Die({False: 1, True: 0})

    if attack_potency > 0 and enemy_potency <= 0:
        # 1 wins.
        return Die({False: 0, True: 1})

    attack_pool = (
        breakthrough_pool(attack_potency, attack_die) if breakthrough else attack_potency @ attack_die
    )

    enemy_pool = enemy_potency @ enemy_die

    # Recursive calls.
    victory_1 = _combat(attack_potency, enemy_potency-damage)
    victory_2 = _combat(attack_potency-1, enemy_potency)

    # Reroll ties and otherwise figure out how many times each side would win.
    outcome = reduce(lambda a,b: Reroll if a == b else a > b, [attack_pool, enemy_pool])

    # Drill down into the inner calls.
    return (outcome).if_else(victory_1, victory_2).simplify()


@cache
def _team_combat(
        leader_potency: int,
        helper_potency: int,
        enemy_potency: int,
        damage_policy: Callable[[int, int], tuple[int, int]],
        inappropriate: bool = False,
        breakthrough: bool = False,
        volunteered: bool = False,
        explode: bool = False,
        n_faces: int = 6
) -> Die:
    """Team combat internal helper.

    # Arguments:
    * `volunteered`: Whether or not the leader's potency was doubled
      by a volunteer. This flag is used to track whether or not the
      leader's dice pool was doubled last round.

    # Returns:
    A Die representing victory or defeat.

    """
    help_die = Die([0 for _ in range(n_faces-1)] + [n_faces])  # Used by non-leaders when teaming up.
    # Maybe use exploding dice:
    die = d(n_faces).explode() if explode else d(n_faces)
    # Inappropriate Cliché: see Risus page 2.
    damage = 3 if inappropriate else 1
    # Voluntarily suffer the loss: see Risus page 3.
    volunteer_potency = 2*leader_potency if volunteered else leader_potency
    # Boxcars and breakthroughs: see Risus Companion page 54.
    leader_pool = breakthrough_pool(volunteer_potency, die) if breakthrough else volunteer_potency@die

    helper_pool = helper_potency @ help_die
    team_pool = leader_pool + helper_pool
    enemy_pool = enemy_potency @ die

    # Base cases:
    if leader_potency > 0 and enemy_potency <= 0:
        # Team victory!
        return Die({True: 1, False: 0})

    if leader_potency <= 0 and enemy_potency > 0:
        # Enemy victory!
        return Die({True: 0, False: 1})

    # Compute outcome and results of combat.
    outcome = reduce(lambda a,b: Reroll if a == b else a > b, [team_pool, enemy_pool])

    damaged_leader, damaged_helper = damage_policy(leader_potency, helper_potency)

    if damaged_leader == leader_potency - 2 or damaged_helper == helper_potency - 2:
        volunteered = True
    else:
        volunteered = False

    team_victory = _team_combat(leader_potency, helper_potency, enemy_potency-damage,
                                        damage_policy,
                                        inappropriate=inappropriate,
                                        breakthrough=breakthrough,
                                        volunteered=volunteered,
                                        explode=explode)

    enemy_victory = _team_combat(damaged_leader, damaged_helper, enemy_potency,
                                         damage_policy,
                                         inappropriate=inappropriate,
                                         breakthrough=breakthrough,
                                         volunteered=volunteered,
                                         explode=explode)

    return outcome.if_else(team_victory, enemy_victory).simplify()


@cache
def breakthrough_pool(potency: int, die: Die, depth: int = 10) -> Die:
    """Make a dice pool that breaks through."""
    pool = potency @ die
    return pool.explode(lambda val: val == pool.max_outcome(), depth=depth)
